"""Some global settings for parboil."""

from pathlib import Path

from .console import out

CFG_DIR = Path("~/.config/parboil").expanduser()
CFG_FILE = CFG_DIR / "config.json"
TPL_DIR = CFG_DIR / "templates"

PRJ_FILE = "parboil.json"
META_FILE = ".parboil"

ERROR_LOG_FILENAME = CFG_DIR / "parboil-errors.log"

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "%(asctime)s:%(name)s:%(process)d:%(lineno)d "
            "%(levelname)s %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
        "simple": {
            "format": "%(message)s",
        },
    },
    "handlers": {
        "logfile": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "ERROR",
            "filename": ERROR_LOG_FILENAME,
            "formatter": "default",
            "backupCount": 2,
        },
        "verbose_output": {
            # "class": "logging.StreamHandler",
            # "stream": "ext://sys.stdout",
            "class": "rich.logging.RichHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "rich_tracebacks": True,
            "tracebacks_suppress": ["click"],
            "console": out,
            "show_time": False,
        },
    },
    "loggers": {
        "parboil": {
            "level": "WARN",
            "handlers": [
                "verbose_output",
            ],
        },
    },
    "root": {"level": "INFO", "handlers": ["logfile"]},
}

if __name__ == "__main__":
    import tomli
    from rich import inspect

    with open(
        "/Users/jneug/Projekte/Skripte/parboil/tests/templates/docs/parboil.toml", "rb"
    ) as toml:
        config = tomli.load(toml)
        inspect(config)



import os
import typing as t
from pathlib import Path


def load_files(dir: Path) -> t.Generator[Path, None, None]:
    for root, dirs, files in os.walk(dir):
        root_path = Path(root)
        for name in files:
            yield root_path.relative_to(dir) / name


def eval_bool(
    value: t.Any, true_values: t.Sequence[str] = ("yes", "true", "y", "1", "ja", "on")
) -> bool:
    return str(value).lower() in true_values



"""Unified output (and input) helpers."""


import textwrap
import typing as t
from functools import partial

import click
import rich
from rich.console import Console
from rich.prompt import (
    Confirm,
    DefaultType,
    FloatPrompt,
    IntPrompt,
    Prompt,
    PromptBase,
    PromptType,
)
from rich.style import Style
from rich.theme import Theme

THEME = Theme(
    {
        # Decorations
        "info.label": "bright_cyan bold",
        "info": "default",
        "error.label": "bright_red bold",
        "error": "red",
        "warn.label": "orange_red1 bold",
        "warn": "orange3",
        "success.label": "bright_green bold",
        "success": "default",
        "question.label": "yellow bold",
        "question": "default",
        # Custom highlight
        "recipe": "bright_magenta",
        "ingredient": "indian_red bold",
        "path": "cyan italic",
        "keyword": "magenta bold",
        "input": "dark_orange",
        "cmd": "indian_red1 italic",
        # Change some default
        "prompt.default": "indian_red",
        "repr.path": "cyan italic",
        "repr.filename": "bright_cyan italic",
    }
)
DECORATIONS = {
    "error": "X",
    "info": "i",
    "question": "?",
    "success": "✓",
    "warn": "!",
}

out = Console(theme=THEME)


def decoration(decor: str) -> str:
    """Creates a decoration for a message shown to the user."""
    # return f"\[[{decor}.label]{DECORATIONS[decor]}[/{decor}.label]]"
    return f"  [{decor}.label]{DECORATIONS[decor]}[/{decor}.label]  "


def clear():
    """Clear the terminal."""
    out.clear()


def sep():
    """Draw a separator with the full width of the terminal."""
    out.print("┄" * out.size.width, style="gray66")


def printd(
    msg: t.Union[str, t.List[str]], decor: str = "info", indent: int = 4
) -> None:
    if isinstance(msg, str):
        msg = msg.split("\n")
    else:
        msg = msg.copy()

    tab = " " * max(0, indent)

    out.print(f"{decoration(decor)} {msg[0]}")
    for line in msg[1:]:
        out.print(f"{tab}{line}")


info = partial(printd, decor="info")
warn = partial(printd, decor="warn")
error = partial(printd, decor="error")
success = partial(printd, decor="success")


def prompt(
    type: t.Type[PromptType],
    msg: t.Union[str, t.List[str]],
    default: t.Any = ...,
    secret: bool = False,
) -> PromptType:
    """Shows a prompt to the user and returns the next input."""
    if isinstance(msg, str):
        msg = msg.split("\n")
    else:
        msg = msg.copy()
    msg[0] = f"{decoration('question')} [question]{msg[0]}[/]"
    msg[1:] = map(lambda _msg: f"    [question]{_msg}[/]", msg[1:])
    for _msg in msg[:-1]:
        out.print(_msg)

    if type is int:
        return IntPrompt(msg[-1], console=out, password=secret)(default=default or ...)  # type: ignore
    elif type is float:
        return FloatPrompt(msg[-1], console=out, password=secret)(default=default or ...)  # type: ignore
    elif type is bool:
        return Confirm(msg[-1], console=out, password=secret)(default=default)  # type: ignore
    else:
        return Prompt(msg[-1], console=out, password=secret)(default=default or ...)  # type: ignore


def question(msg: t.Union[str, t.List[str]], key: str, default: t.Any = ...) -> str:
    if isinstance(msg, str):
        msg = msg.split("\n")
    msg.append(f"[field]{key}[/field]")

    return prompt(str, msg, default=default)


def question_int(msg: t.Union[str, t.List[str]], key: str, default: t.Any = ...) -> int:
    if isinstance(msg, str):
        msg = msg.split("\n")
    msg.append(f"[field]{key}[/field]")

    return prompt(int, msg, default=default)


def confirm(msg: t.Union[str, t.List[str]], default: t.Any = ...) -> bool:
    return bool(prompt(bool, msg, default=bool(default)))


def choice(
    msg: t.Union[str, t.List[str]],
    choices: t.Sequence[PromptType],
    default: t.Union[PromptType, int, None] = None,
) -> t.Tuple[int, PromptType]:
    if isinstance(msg, str):
        msg = msg.split("\n")
    for i, text in enumerate(choices):
        msg.append(f"[input]{i+1}[/] - [keyword]{text}[/]")
    msg.append(f"Select from [input]1..{len(choices)}[/]")

    # get correct index for default answer
    if isinstance(default, str):
        try:
            i = choices.index(default)
        except ValueError:
            default = None
        else:
            default = i + 1
    elif isinstance(default, int) and not (0 <= default < len(choices)):
        default = None

    while True:
        answer = prompt(int, msg, default=default or ...)
        if not (0 <= (answer - 1) < len(choices)):
            warn(
                f"{answer} is not a valid choice. Please select from range [input]1..{len(choices)}[/]."
            )
        else:
            return answer, choices[answer - 1]



# -*- coding: utf-8 -*-

import os
import platform
import re
import string
import typing as t
import unicodedata
from datetime import datetime
from functools import update_wrapper

from click import pass_context
from jinja2 import nodes
from jinja2.ext import Environment, Extension
from jinja2.parser import Parser


def pass_tpldir(f: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
    @pass_context
    def new_func(ctx, *args, **kwargs):
        if ctx.obj and "TPLDIR" in ctx.obj:
            return ctx.invoke(f, ctx.obj["TPLDIR"], *args, **kwargs)
        else:
            return ctx.invoke(f, ctx.obj["TPLDIR"], *args, **kwargs)

    return update_wrapper(new_func, f)


DEFAULT_REPLACE = dict(ä="ae", ö="oe", ü="ue", ß="ss")


def jinja_filter_fileify(
    s: t.Any,
    sep: str = "_",
    replace: t.Dict[str, str] = DEFAULT_REPLACE,
    char_limit: int = 88,
) -> str:
    """
    Django util.text.get_valid_filename
    """
    # s = str(s).strip().replace(" ", "_")
    # return re.sub(r"(?u)[^-\w.]", "", s)

    # https://gist.github.com/wassname/1393c4a57cfcbf03641dbc31886123b8
    # replace spaces
    s = str(s).strip().replace(" ", sep)
    for k, v in replace.items():
        s = s.replace(k, v)

    # keep only valid ascii chars
    s = unicodedata.normalize("NFKD", s).encode("ASCII", "ignore").decode()

    if "windows" in platform.system().lower():
        if len(s) > 255:
            s, ext = os.path.splitext(s)
            return s[: (char_limit - len(ext))] + ext
    return s


def jinja_filter_slugify(value: t.Any, allow_unicode: bool = False) -> str:
    """
    Django util.text.slugify
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def jinja_filter_roman(value: t.Any) -> str:
    """Convert an integer to a Roman numeral.
    https://www.oreilly.com/library/view/python-cookbook/0596001673/ch03s24.html
    """

    # if not isinstance(value, type(1)):
    if not isinstance(value, int):
        return str(value)
    if not 0 < value < 4000:
        return str(value)
    ints = (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1)
    nums = ("M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I")
    result = []
    for i in range(len(ints)):
        count = int(value / ints[i])
        result.append(nums[i] * count)
        value -= ints[i] * count
    return "".join(result)



# -*- coding: utf-8 -*-

import os
import shlex
import subprocess
import typing as t
from dataclasses import dataclass, field

from jinja2 import Environment
from rich.live import Live
from rich.panel import Panel
from rich.text import Text


@dataclass
class Task:
    """Stores and esecutes commands to run before or after compiling a recipe."""

    cmd: t.Union[str, t.List[str]]
    env: t.Optional[dict] = None
    quiet: bool = False

    returncode: int = field(init=False, default=-1)

    _shell: bool = field(default=False, init=False)

    def __post_init__(self):
        # make sure command is a list
        if isinstance(self.cmd, str):
            self._shell = True
            self.cmd = [self.cmd]
            # self.cmd = shlex.split(self.cmd)

    @classmethod
    def from_dict(self, descr: t.Dict[str, t.Any]) -> "Task":
        for k in list(descr.keys()):
            # everything not in instance variables is removed
            if k not in Task.__dict__["__annotations__"].keys():
                del descr[k]
        return Task(**descr)

    def __templates__(self) -> t.Generator[str, str, None]:
        if isinstance(self.cmd, str):
            self.cmd = yield self.cmd
        else:
            for i, c in enumerate(self.cmd):
                self.cmd[i] = yield c

    def execute(self) -> bool:
        if self.env:
            environ = os.environ.copy().update(self.env)
        else:
            environ = os.environ.copy()

        result = subprocess.run(
            self.cmd,
            shell=self._shell,
            # check=True,
            env=environ,
            stdout=subprocess.DEVNULL if self.quiet else None,
            stderr=subprocess.STDOUT,
        )
        self.returncode = result.returncode
        return result.returncode == 0

    def quoted(self):
        return " ".join(shlex.quote(c) for c in self.cmd)

    def __str__(self):
        return self.quoted()



# -*- coding: utf-8 -*-


"""
Parboil lets you generate boilerplate projects from recipe files.

Run boil --help for more info.
"""

import json
import logging.config
import os
import platform
import re
import shutil
import subprocess
import time
import typing as t
from pathlib import Path

import click
import jsonc
import rich
from colorama import Back, Fore, Style
from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PrefixLoader
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.tree import Tree

import parboil.console as console
from parboil import __version__

from .errors import ProjectError, ProjectExistsError, ProjectFileNotFoundError
from .ext import pass_tpldir
from .recipes import Boiler, Recipe, Repository
from .settings import CFG_DIR, CFG_FILE, LOGGING_CONFIG, TPL_DIR

logger = logging.getLogger("parboil")

USE_MARKUP = dict(markup=True)
USE_MARKUP_NO_HIGHLIGHT = dict(markup=True, highlight=False)


@click.group()
@click.version_option(version=__version__, prog_name="parboil")
@click.option(
    "-c",
    "--config",
    type=click.File(),
    envvar="BOIL_CONFIG",
    help="Provides a different json config file for this run. Read from stdin with -.",
)
@click.option(
    "--repo",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    envvar="BOIL_REPO",
    help="Location of the local recipe repository.",
)
@click.option("--debug", is_flag=True)
@click.pass_context
def boil(
    ctx: click.Context,
    config: t.TextIO,
    repo: Path,
    debug: bool = False,
) -> None:
    ctx.ensure_object(dict)

    # Setup logging
    logging.config.dictConfig(LOGGING_CONFIG)
    if debug:
        logger.setLevel(logging.DEBUG)

    logger.info(
        "Starting up [b]parboil[/] :rice:, version [bright_cyan bold]%s[/] (Python [bright_cyan bold]%s[/])",
        __version__,
        platform.python_version(),
        extra=USE_MARKUP,
    )

    # Load config file
    if config:
        try:
            user_cfg = jsonc.load(config)
            ctx.obj = {**ctx.obj, **user_cfg}
            logger.info("Merged in config from %s", config)
        except json.JSONDecodeError:
            logger.warn("Error loading config from %s", config)
    else:
        if CFG_FILE.exists():
            with open(CFG_FILE) as f:
                try:
                    cmd_cfg = json.load(f)
                    ctx.obj = {**ctx.obj, **cmd_cfg}
                    logger.info("Merged in config from %s", str(CFG_FILE))
                except json.JSONDecodeError:
                    logger.warn("Error loading config from %s", str(CFG_FILE))

    ctx.obj["TPLDIR"] = repo or TPL_DIR
    logger.info("Working with recipe repository %s", str(ctx.obj["TPLDIR"]))


@boil.command(short_help="List installed recipes")
@click.option("-p", "--plain", is_flag=True)
@pass_tpldir
def list(TPLDIR: Path, plain: bool) -> None:
    """
    Lists all recipes in the active local repository.
    """
    repo = Repository(TPLDIR)
    if repo.exists():
        if len(repo) > 0:
            if plain:
                for project_name in repo:
                    console.out.print(project_name)
            else:
                table = Table(
                    title=f"Recipes installed in [path]{TPLDIR}[/path]",
                    # expand=True,
                    box=rich.box.MINIMAL_DOUBLE_HEAD,
                    # show_lines=True,
                )

                table.add_column("Name", style="keyword")
                table.add_column(
                    "[purple]Created[/] / [purple]Updated[/] [bright_black]or[/] [path]Realpath[/]"
                )

                for recipe in sorted(repo.recipes(), key=lambda t: t.name):
                    recipe.load()

                    name = recipe.name
                    created = "[white on red]unknown[/]"
                    updated = "[bright_black]never[/]"
                    if recipe.is_symlinked():
                        name = f"[cyan]{recipe.name}*[/]"
                        data = f"[path]{recipe.root}[/]"
                    else:
                        if "updated" in recipe.meta:
                            updated = time.ctime(int(recipe.meta["updated"]))
                        if "created" in recipe.meta:
                            created = time.ctime(int(recipe.meta["created"]))
                        data = f"[purple]{created}[/] / [purple]{updated}[/]"

                    table.add_row(name, data)

                console.out.print(table)
        else:
            console.info("No recipes installed yet.")
    else:
        console.warn("Repository directory does not exist.")


@boil.command(short_help="Install a new project recipe")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Set this flag to overwrite existing recipes named RECIPE without prompting.",
)
@click.option(
    "-d",
    "--download",
    is_flag=True,
    help="Set this flag if SOURCE is a github repository to download instead of a local directory.",
)
@click.option("-r", "--repo", "is_repo", is_flag=True)
@click.option("-s", "--symlink", "symlink", is_flag=True)
@click.argument("source")
@click.argument("recipe", required=False)
@click.pass_context
def install(
    ctx: click.Context,
    source: str,
    recipe: str,
    force: bool,
    download: bool,
    is_repo: bool,
    symlink: bool,
) -> None:
    """
    Install a recipe named RECIPE from SOURCE to the local recipe repository.

    SOURCE may be a local directory or the url of a GitHub repository. You may also pass in the name of a repository in the form user/repo, but need to set the -d flag to indicate it isn't a local directory.

    -r indicates that SOURCE is a folder with multiple recipes that should be installed.

    Use -s to create symlinks instead of copying the files. (Useful for recipe development.)
    """
    # logger = logging.getLogger("parboil")

    # TODO: validate recipes!
    TPLDIR = ctx.obj["TPLDIR"]
    repo = Repository(TPLDIR)

    # is source a github url? Then assume -d
    if re.match(r"https?://(www\.)?github\.com", source):
        download = True
    # set missing arguments
    if download:
        if re.match("[A-Za-z_-]+/[A-Za-z_-]+", source):
            source = f"https://github.com/{source}"
        if not recipe:
            recipe = source.split("/")[-1]
    else:
        if not recipe:
            recipe = Path(source).name

    if not is_repo and not force and repo.is_installed(recipe):
        if not console.confirm(f"Overwrite existing recipe named [recipe]{recipe}[/]?"):
            ctx.abort()

    try:
        if download:
            projects = repo.install_from_github(
                recipe, source, hard=True, is_repo=is_repo
            )
        else:
            projects = repo.install_from_directory(
                recipe, source, hard=True, is_repo=is_repo, symlink=symlink
            )
    except ProjectError as fnfe:
        console.error(str(fnfe))
    except FileExistsError as fee:
        console.error(str(fee))
    except shutil.Error:
        console.error(f"Could not install recipe [recipe]{recipe}[/]")
    else:
        if not projects:
            console.success("No recipes where installed.")
            return

        for project in projects:
            console.success(f"Installed recipe [recipe]{project.name}[/]")
        if len(projects) == 1:
            console.printd(f"\nUse with [cmd]boil use {projects[0].name}[/]")
        else:
            console.printd("\nUse with [cmd]boil use <recipe_name>[/]")


@boil.command(short_help="Uninstall an existing recipe")
@click.option("-f", "--force", is_flag=True)
@click.argument("recipe")
@pass_tpldir
def uninstall(TPLDIR: Path, force: bool, recipe: str) -> None:
    repo = Repository(TPLDIR)

    if repo.is_installed(recipe):
        rm = force
        if not force:
            rm = console.confirm(
                f"Do you really want to uninstall recipe [recipe]{recipe}[/]"
            )
        if rm:
            try:
                repo.uninstall(recipe)
                console.success(f"Removed recipe [recipe]{recipe}[/]")
            except OSError:
                console.error(
                    [
                        f"Error while uninstalling recipe [recipe]{recipe}[/]",
                        "You might need to manually delete the recipe directory at",
                        f"[path]{repo.root}[/]",
                    ]
                )
    else:
        console.warn(f"Recipe [recipe]{recipe}[/] does not exist")


@boil.command(short_help="Update an existing recipe")
@click.argument("recipe")
@click.pass_context
def update(ctx: click.Context, recipe: str) -> None:
    """
    Update RECIPE from the source it was first installed from.
    """
    cfg = ctx.obj

    repo = Repository(cfg["TPLDIR"])

    if not repo.is_installed(recipe):
        console.error(f"Recipe [recipe]{recipe}[/] does not exist.")
        ctx.exit(2)

    _recipe = repo.get_recipe(recipe, load=True)
    try:
        repo.update(_recipe)
    except ProjectFileNotFoundError as pe:
        console.error(
            [
                str(pe),
                "To update templates make sure to install with [cmd]boil install[/].",
            ]
        )
        ctx.abort()
    except ProjectError as pe:
        console.error(str(pe))
        ctx.abort()
    else:
        if _recipe.meta["source_type"] == "github":
            console.success(f"Updated template [recipe]{recipe}[/] from GitHub.")
        else:
            console.success(
                f"Updated template [recipe]{recipe}[/] from local filesystem."
            )


@boil.command(short_help="Use an existing recipe")
@click.option(
    "--hard",
    is_flag=True,
    help="Force overwrite of existing output directory. If the directory OUT exists and is not empty, it will be deleted and newly created.",
)
@click.option(
    "-v",
    "--value",
    multiple=True,
    nargs=2,
    help="Sets a prefilled value for the recipe.",
)
@click.option("--dev", is_flag=True)
@click.argument("recipe")
@click.argument(
    "out", default=".", type=click.Path(file_okay=False, dir_okay=True, writable=True)
)
@click.pass_context
def use(
    ctx: click.Context,
    recipe: str,
    out: t.Union[str, Path],
    hard: bool,
    value: t.List[t.Tuple[str, str]],
    dev: bool = False,
) -> None:
    """
    Generate a new project from RECIPE.

    If OUT is given and a directory, the recipe is created there.
    Otherwise the cwd is used.
    """
    cfg = ctx.obj

    # Check teamplate and read configuration
    repo = Repository(cfg["TPLDIR"])
    _recipe = repo.get_recipe(recipe)

    try:
        _recipe.load()
    except FileNotFoundError:
        console.warn(f"No valid recipe found for name [recipe]{recipe}[/]")
        ctx.exit(1)

    # Prepare output directory
    # if out == ".":
    #     out = Path.cwd()
    # else:
    #     out = Path(out)
    out = Path(out).resolve()

    if out.exists() and len(os.listdir(out)) > 0:
        if hard:
            shutil.rmtree(out)
            out.mkdir(parents=True)
            console.success(f"Cleared [path]{out}[/]")
    elif not out.exists():
        out.mkdir(parents=True)
        console.success(f"Created [path]{out}[/]")

    ## Prepare prefilled values
    prefilled = cfg.get("prefilled", dict())
    for key, val in value:
        prefilled[key] = val

    ## Prepare project and read user answers
    project = Boiler(_recipe, out, prefilled)
    project.fill()

    for success, file_in, file_out in project.compile():
        if success:
            console.success(f"Created [path]{file_out}[/]")
        else:
            console.warn(f"Skipped [path]{file_out}[/] due to empty content")

    console.success(
        f'Generated project for recipe "[recipe]{_recipe.name}[/]" in [path]{out}[/]'
    )


@boil.command(short_help="Show information about an installed recipe")
@click.option(
    "--conf",
    is_flag=True,
    help="Print the full project file.",
)
@click.option(
    "--tree",
    is_flag=True,
    help="Print the full recipe tree.",
)
@click.argument("recipe")
@click.pass_context
def info(ctx: click.Context, recipe: str, conf: bool, tree: bool) -> None:
    cfg = ctx.obj

    repo = Repository(cfg["TPLDIR"])
    if not repo.is_installed(recipe):
        console.warn(f"No recipe [recipe]{recipe}[/] installed.")
        return

    _recipe = repo.get_recipe(recipe)

    info_table = Table(
        title=f"Information for recipe [recipe]{_recipe.name}[/]",
        highlight=True,
        box=rich.box.MINIMAL_DOUBLE_HEAD,
        # show_lines=True,
    )

    info_table.add_column("Property", style="keyword")
    info_table.add_column("Value")

    info_table.add_row("path", str(_recipe.root))
    info_table.add_row("is symlinked", str(_recipe.is_symlinked()))

    console.out.print(info_table)

    if tree:
        _tree = Tree(_recipe.name)
        _walk_directory(_recipe.root, _tree)
        _tree_panel = Panel(_tree, title=f"Contents of {_recipe.root}")
        console.out.print(_tree_panel)

    if conf:
        with open(_recipe.recipe_file, "rt") as cf:
            _syntax = Syntax(cf.read(), lexer="json")
            _syntax_panel = Panel(_syntax, title=str(_recipe.recipe_file))
            console.out.print(_syntax_panel)


def _walk_directory(directory: Path, tree: Tree) -> None:
    """Recursively build a Tree with directory contents."""
    from rich.filesize import decimal
    from rich.markup import escape

    # Sort dirs first then by filename
    paths = sorted(
        Path(directory).iterdir(),
        key=lambda path: (path.is_file(), path.name.lower()),
    )

    for path in paths:
        # Remove hidden files
        if path.name.startswith("."):
            continue
        if path.is_dir():
            style = "dim" if path.name.startswith("__") else ""
            branch = tree.add(
                f"[bold magenta]:open_file_folder: [link file://{path}]{escape(path.name)}",
                style=style,
                guide_style=style,
            )
            _walk_directory(path, branch)
        else:
            text_filename = rich.text.Text(path.name, "green")
            text_filename.highlight_regex(r"\..*$", "bold red")
            text_filename.stylize(f"link file://{path}")
            file_size = path.stat().st_size
            text_filename.append(f" ({decimal(file_size)})", "blue")
            icon = "🐍 " if path.suffix == ".py" else "📄 "
            tree.add(rich.text.Text(icon) + text_filename)



# -*- coding: utf-8 -*-

import sys
import typing as t
from collections.abc import MutableMapping
from dataclasses import dataclass, field

import click
import rich
import rich.prompt
from colorama import Back, Fore, Style
from rich import inspect

import parboil.console as console

from .errors import ProjectConfigError

if t.TYPE_CHECKING:
    from parboil.recipes import Boiler, Recipe

VTYPE = t.TypeVar("VTYPE")

OptStr = t.Optional[str]
OptVtype = t.Optional[VTYPE]


MSG_DEFAULT = 'Enter a value for "[ingredient]{{INGREDIENT.name}}[/]"'
MSG_CHOICE = 'Chose a value for "[ingredient]{{INGREDIENT.name}}[/]"'
MSG_DISABLE = 'Do you want do [italic]disable[/] "[ingredient]{{INGREDIENT.name}}[/]"'
MSG_ENABLE = 'Do you want do [italic]enable[/] "[ingredient]{{INGREDIENT.name}}[/]"'
MSG_CONFIRM = 'Do you want do [italic]enable[/] "[ingredient]{{INGREDIENT.name}}[/]"'


def get_ingredient(name: str, definition: t.Dict[t.Any, t.Any]) -> "Ingredient":
    global INGREDIENT_TYPES

    # deal with shorthand definition
    if isinstance(definition, list):
        definition = dict(field_type="choice", choices=definition)
    elif not isinstance(definition, dict):
        definition = dict(default=definition)

    ## determine field type
    field_type = definition.get("field_type", "default")
    if "field_type" in definition:
        del definition["field_type"]

    ## Special case subtemplates
    if field_type == "recipe":
        definition["recipe_name"] = definition["name"]

    # name key is reserved
    if "name" in definition:
        del definition["name"]

    # select proper field instance based on default value
    if field_type == "default":
        _d = definition.get("default", None)
        _c = definition.get("choices", None)
        if isinstance(_d, bool):
            field_type = "confirm"
        if isinstance(_d, int):
            definition["type"] = "int"
        elif isinstance(_c, list):
            field_type = "choice"
        elif isinstance(_c, dict):
            field_type = "dict"

    # create field instance
    try:
        return INGREDIENT_TYPES[field_type](name, **definition)
    except NameError:
        raise ProjectConfigError(f"Unknown field type {field_type}.")


@dataclass(init=False)
class Ingredient(t.Generic[VTYPE]):
    name: str
    value: OptVtype = None
    default: OptVtype = None
    help: str = MSG_DEFAULT
    condition: OptStr = None
    optional: bool = False

    ## Holds arbitrary arguments passed to __init__
    args: t.Dict[str, t.Any] = field(default_factory=dict)

    type: str = "str"

    def __init__(
        self,
        name: str,
        value: OptVtype = None,
        default: OptVtype = None,
        help: OptStr = None,
        condition: OptStr = None,
        optional: bool = False,
        type: str = "str",
        **kwargs,
    ):
        self.name = name

        self.default = default
        self._value = value
        if help:
            self.help = help
        self.condition = condition
        self.optional = optional

        self.args = dict()
        if kwargs:
            self.args.update(kwargs)

        self.type = type

    def __templates__(self) -> t.Generator[str, str, None]:
        for key in ["help", "default", "value", "condition"]:
            val = getattr(self, key, None)
            if val is not None and isinstance(val, str):
                setattr(self, key, (yield val))

    def prompt(self, boiler: "Boiler") -> OptVtype:
        """
        Prompts the user for an answer and stores the result in self.value.

        If the field already has a value, no prompt is shown. To force a prompt
        call `del field.value` first.
        """
        if not self.value:
            self._prompt(boiler)
        return self.value

    def _prompt(self, boiler: "Boiler") -> None:
        """Actually prompt the user for an answer and store the
        result in `self.value`."""
        if self.type == "int":
            self.value = console.question_int(
                self.help,
                key=self.name,
                default=self.default,
            )
        else:
            self.value = console.question(
                self.help,
                key=self.name,
                default=self.default,
            )


class ConfirmIngredient(Ingredient):
    def __init__(self, name, **kwargs):
        # if bool(getattr(kwargs, "default", False)):
        #     self.help = MSG_DISABLE
        # else:
        #     self.help = MSG_ENABLE
        self.help = MSG_CONFIRM
        super().__init__(name, **kwargs)

    def _prompt(self, boiler: "Boiler") -> None:
        self.value = console.confirm(
            self.help,
            default=bool(self.default),
        )
        # if bool(self.default):
        #     self.value = not console.confirm(
        #         self.help,
        #         default=True,
        #     )
        # else:
        #     self.value = console.confirm(
        #         self.help,
        #         default=False,
        #     )


class ChoiceIngredient(Ingredient):
    _choices: t.List[str] = field(default_factory=list)

    def __init__(self, name, choices: t.List[t.Any], **kwargs):
        self.help = MSG_CHOICE
        self._choices = choices.copy()
        super().__init__(name, **kwargs)

    def __templates__(self) -> t.Generator[str, str, None]:
        yield from super().__templates__()
        for i, choice in enumerate(self.choices):
            self._choices[i] = yield choice

    @property
    def choices(self) -> t.List[str]:
        return self._choices

    def _prompt(self, boiler: "Boiler") -> None:
        if len(self.choices) > 1:
            index, self.value = console.choice(
                self.help,
                choices=self.choices,
                default=self.default,
            )
            boiler.context[f"{self.name}_index"] = index
        elif len(self.choices) == 1:
            self.value = self.choices[0]
            boiler.context[f"{self.name}_index"] = 0
        else:
            self.value = None


class FileselectIngredient(ChoiceIngredient):
    def _prompt(self, boiler: "Boiler") -> None:
        super()._prompt(boiler)

        if self.value:
            boiler.recipe.templates.append(f"includes:{self.value}")
            # optionally update file config with filename
            if "filename" in self.args:
                file_config = getattr(boiler.recipe.files, self.value, dict())
                file_config.update({"filename": self.args["filename"]})
                boiler.recipe.files[self.value] = file_config


class ChoiceDictIngredient(ChoiceIngredient):
    _values: t.List[str] = field(default_factory=list)

    def __init__(self, name, choices: t.Dict[t.Any, str], **kwargs):
        self.help = MSG_CHOICE
        self._values = list(choices.values())
        super().__init__(name, choices=list(choices.keys()), **kwargs)

    def __templates__(self) -> t.Generator[str, str, None]:
        super().__templates__()
        for i, val in enumerate(self._values):
            self._values[i] = yield val

    def _prompt(self, boiler: "Boiler") -> None:
        super()._prompt(boiler)

        if self.value:
            i = boiler.context[f"{self.name}_index"]
            boiler.context[f"{self.name}_key"] = self.value
            self.value = self._values[i]


class RecipeIngredient(Ingredient):
    _recipe_name: str = ""

    def __init__(self, name, recipe_name: str, **kwargs):
        self._recipe_name = recipe_name
        super().__init__(name, **kwargs)

    def __templates__(self) -> t.Generator[str, str, None]:
        super().__templates__()
        self._recipe_name = yield self._recipe_name

    @property
    def recipe_name(self) -> str:
        return self._recipe_name

    def _prompt(self, boiler: "Boiler") -> None:
        console.info(f'Including subrecipe "[recipe]{self.recipe_name}[/]"')

        if boiler.recipe.repository:
            subrecipe = boiler.recipe.repository.get_recipe(self._recipe_name)
        else:
            try:
                Recipe.__name__
            except NameError:
                from .recipes import Recipe

            subrecipe = Recipe(self._recipe_name, boiler.recipe.root.parent)

        if subrecipe.is_valid():
            prefilled = boiler.prefilled.copy()
            if "pass_context" not in self.args or self.args["pass_context"] is True:
                prefilled.update(boiler.context)

            try:
                Boiler.__name__
            except NameError:
                from .recipes import Boiler

            subrecipe.load()
            subboiler = Boiler(subrecipe, boiler.target_dir, prefilled)
            subboiler.fill()
            boiler.recipe.templates.append(subrecipe)
            boiler.context.maps.append({self.name: subrecipe.context})


INGREDIENT_TYPES = {
    "default": Ingredient,
    "confirm": ConfirmIngredient,
    "choice": ChoiceIngredient,
    "file_select": FileselectIngredient,
    "dict": ChoiceDictIngredient,
    "recipe": RecipeIngredient,
}



# -*- coding: utf-8 -*-
"""Handling of recipes.

Classes for handling recipes (boilerplate templates), repositories
(directories with installed templates) and boilers (projects that compile
templates into a specific target directory).
"""


import json
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
import typing as t
from collections import ChainMap
from collections.abc import Mapping, MutableMapping, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path

import click
import jsonc
from rich import inspect

import parboil.console as console

from .errors import (
    ProjectError,
    ProjectExistsError,
    ProjectFileNotFoundError,
    RecipeNotInstalledError,
    TaskExecutionError,
    TaskFailedError,
)
from .helpers import eval_bool, load_files
from .ingredients import Ingredient, get_ingredient
from .renderer import ParboilRenderer
from .settings import META_FILE, PRJ_FILE
from .tasks import Task

logger = logging.getLogger(__name__)

RESERVED_KEYS = ("_tasks", "_files", "_context", "_settings")


@dataclass(init=False)
class Recipe:
    """Container for recipe information.

    Recipes hold information about a template that can be used by parboil
    to generate a project with user answers. The compilation process is handled in
    a [parboil.recipes.Boiler][] instance.

    Raises:
        ProjectFileNotFoundError: In case some mandatory recipe files are missing.
        ProjectError: Any error related to project initialization.
    """

    name: str

    repository: "Repository"
    _root: Path

    recipe_file: Path
    meta_file: Path
    templates_dir: Path
    includes_dir: Path

    meta: t.Dict[str, t.Any] = field(default_factory=dict)
    files: t.Dict[str, t.Dict[str, t.Any]] = field(default_factory=dict)
    templates: t.List[t.Union[str, Path, "Recipe"]] = field(default_factory=list)
    includes: t.List[Path] = field(default_factory=list)

    ingredients: t.List[Ingredient] = field(default_factory=list)
    context: t.ChainMap[str, t.Any] = field(default_factory=ChainMap)

    tasks: t.Dict[str, t.List[Task]] = field(default_factory=dict)

    def __init__(
        self,
        name: str,
        repository: t.Union[str, Path, "Repository"],
        load: bool = False,
    ):
        self.name = name
        if isinstance(repository, Repository):
            self.repository = repository
        else:
            self.repository = Repository(repository)
        self._root = self.repository.root / name

        # setup config files and paths
        self.recipe_file = self.root / PRJ_FILE
        self.meta_file = self.root / META_FILE
        self.templates_dir = self.root / "template"
        self.includes_dir = self.root / "includes"

        self.meta = dict()
        self.files = dict()
        self.templates = list()
        self.includes = list()
        self.ingredients = list()
        self.context = ChainMap()
        self.tasks = {"pre-run": [], "post-run": []}

        if load:
            self.load()

    @property
    def root(self) -> Path:
        if self.is_symlinked():
            return self._root.resolve()
        else:
            return self._root

    def is_symlinked(self) -> bool:
        return self._root.is_symlink()

    def exists(self) -> bool:
        return self._root.is_dir()

    def is_valid(self) -> bool:
        return self.exists() and self.recipe_file.is_file()

    def load(self) -> None:
        """Loads the project file and some metadata."""
        # Load files form template folder
        self.templates.extend(load_files(self.templates_dir))
        self.includes.extend(load_files(self.includes_dir))

        # Load config
        config: t.Dict[str, t.Any] = dict()
        try:
            with open(self.recipe_file) as f:
                config = jsonc.load(f)
        except FileNotFoundError as e:
            raise ProjectFileNotFoundError() from e
        except json.JSONDecodeError as e:
            raise ProjectError("Malformed project file.") from e

        if "_files" in config:
            for file, data in config["_files"].items():
                if isinstance(data, str):
                    self.files[file] = dict(filename=data)
                else:
                    self.files[file] = data

        # Parse config
        self._load_ingredients(config)
        self._load_tasks(config)

        if "_context" in config:
            self.context.maps.append({**config["_context"]})

        # Load metafile
        if self.meta_file.is_file():
            with open(self.meta_file) as f:
                self.meta = {**self.meta, **json.load(f)}

    def _load_ingredients(self, config: t.Dict[str, t.Any]) -> None:
        """
        Parse `fields` key from `config` into `Ingredient` objects
        and stores them in the `ingredients` attribute.
        """
        for k, v in config.items():
            if k not in RESERVED_KEYS:
                self.ingredients.append(get_ingredient(k, v))

    def _load_tasks(self, config: t.Dict[str, t.Any]) -> None:
        """Parse ``tasks`` key from ``config`` into :class:`Task` objects and stores them in the ``tasks`` attribute."""
        if "_tasks" in config:
            for hook in self.tasks.keys():
                if hook in config["_tasks"]:
                    for task_def in config["_tasks"][hook]:
                        if isinstance(task_def, str) or isinstance(task_def, list):
                            self.tasks[hook].append(Task(task_def))
                        elif isinstance(task_def, dict):
                            self.tasks[hook].append(Task(**task_def))

    def save(self) -> None:
        """Saves the current meta file to disk."""
        if self.meta_file:
            with open(self.meta_file, "w") as f:
                json.dump(self.meta, f)


class Repository(Mapping[str, Recipe]):
    def __init__(self, root: t.Union[str, Path]) -> None:
        self._root: Path = Path(root)
        self._recipes: t.List[str] = list()
        self.load()

    @property
    def root(self) -> Path:
        return self._root

    def exists(self) -> bool:
        return self._root.is_dir()

    def load(self):
        logger.info("Loading repository from `%s`", self._root)
        ## Remove previously loaded templates
        self._recipes = list()
        if self.exists():
            for child in self._root.iterdir():
                if child.is_dir():
                    project_file = child / PRJ_FILE
                    if project_file.is_file():
                        self._recipes.append(child.name)
                        logger.debug("---> %s", child.name)

    def __len__(self) -> int:
        return len(self._recipes)

    def __iter__(self) -> t.Generator[str, None, None]:
        yield from self._recipes

    def __getitem__(self, name) -> Recipe:
        return self.get_recipe(name)

    def is_installed(self, recipe: str) -> bool:
        # TODO is thos enough?
        recipe_dir = self._root / recipe
        return recipe_dir.is_dir()

    def recipes(self) -> t.Generator[Recipe, None, None]:
        yield from (self.get_recipe(name) for name in self)

    def get_recipe(self, recipe: str, load: bool = False) -> Recipe:
        r = Recipe(recipe, self)
        if load:
            r.load()
        return r

    def install_from_directory(
        self,
        recipe: str,
        source: t.Union[str, Path],
        hard: bool = False,
        is_repo: bool = False,
        symlink: bool = False,
        reload: bool = True,
    ) -> t.List[Recipe]:
        """
        If source contains a valid recipe it is installed
        into this local repository and a `Recipe` object is returned.
        """
        logger.info(
            f"Starting install from directory {source!s}", extra={"repository": self}
        )

        if self.is_installed(recipe):
            if not hard:
                raise ProjectExistsError(
                    "The template already exists. Delete first or retry install with hard=True."
                )
            else:
                self._delete(recipe)
                logger.debug(f"Deleted existing template {recipe}")

        ## check source directory
        source = Path(source).resolve()
        if not source.is_dir():
            raise ProjectFileNotFoundError("Source does not exist.")

        if not is_repo:
            logger.debug(
                "Attempting to install from source %s",
                source,
                extra={"repository": self},
            )

            project_file = source / PRJ_FILE
            template_dir = source / "template"

            if not project_file.is_file():
                raise ProjectFileNotFoundError(
                    f"The source does not contain a {PRJ_FILE} file."
                )

            if not template_dir.is_dir():
                raise ProjectFileNotFoundError(
                    "The source does not contain a template directory."
                )

            # install template
            if not symlink:
                # copy full template tree
                shutil.copytree(source, self._root / recipe)

                # create meta file
                _template = self.get_recipe(recipe)
                _template.meta = {
                    "created": time.time(),
                    "source_type": "local",
                    "source": str(source),
                }
                _template.save()
            else:
                # create a symlink
                os.symlink(source, self._root / recipe, target_is_directory=True)
                _template = self.get_recipe(recipe)

            templates = [_template]
        else:
            templates = list()

            for child in source.iterdir():
                if child.is_dir():
                    logger.debug(
                        "Attempting to install from subfolder %s",
                        child,
                        extra={"repository": self},
                    )

                    project_file = child / PRJ_FILE
                    if project_file.is_file():
                        try:
                            _template = self.install_from_directory(
                                child.name, child, hard=hard, reload=False
                            )[0]
                            templates.append(_template)
                        except ProjectFileNotFoundError:
                            logger.warn(
                                "Subfolder %s is not a valid subfolder",
                                child,
                                extra={"repository": self},
                            )
                            pass

        if reload:
            self.load()
        return templates

    def install_from_github(
        self, template: str, url: str, hard: bool = False, is_repo: bool = False
    ) -> t.List[Recipe]:
        if not is_repo:
            # check target dir
            if self.is_installed(template):
                if not hard:
                    raise ProjectExistsError(
                        "The template already exists. Delete first or retry install with hard=True."
                    )
                else:
                    self._delete(template)

            project = self.get_recipe(template)

            # do git clone
            # TODO: Does this work on windows?
            git = subprocess.Popen(["git", "clone", url, str(project.root)])
            git.wait(30)

            # create meta file
            project.meta = {
                "created": time.time(),
                "source_type": "github",
                "source": url,
            }
            project.save()

            self.load()
            return [project]
        else:
            projects = list()  # return list of installed projects

            # do git clone into temp folder
            with tempfile.TemporaryDirectory() as temp_repo:
                git = subprocess.Popen(["git", "clone", url, temp_repo])
                git.wait(30)

                for child in Path(temp_repo).iterdir():
                    if child.is_dir():
                        project_file = child / PRJ_FILE
                        if project_file.is_file():
                            try:
                                project = self.install_from_directory(
                                    child.name, child, hard=hard
                                )[0]
                                # remove source data
                                del project.meta["source_type"]
                                del project.meta["source"]
                                project.save()

                                projects.append(project)
                            except ProjectFileNotFoundError:
                                pass
                            except ProjectExistsError:
                                pass

            self.load()
            return projects

    def uninstall(self, template: str) -> None:
        self._delete(template)

    def update(self, recipe: t.Union[str, Recipe], hard: bool = False) -> None:
        """
        Update an template from its original source.

        Does not work for symlinked templates.
        """
        if isinstance(recipe, str):
            if not self.is_installed(recipe):
                raise RecipeNotInstalledError(recipe, self)
                return
            recipe = self.get_recipe(recipe)

        if not recipe.meta_file.exists():
            raise ProjectFileNotFoundError(
                "Template metafile does not exist. Can't read update information."
            )

        if recipe.meta["source_type"] == "github":
            git = subprocess.Popen(["git", "pull", "--rebase"], cwd=recipe.root)
            git.wait(30)
        elif recipe.meta["source_type"] == "local":
            if Path(recipe.meta["source"]).is_dir():
                shutil.rmtree(recipe.root)
                shutil.copytree(recipe.meta["source"], recipe.root)
            else:
                raise ProjectError("Original source directory no longer exists.")
        else:
            raise ProjectError("No source information found.")

        # Update meta file for later updates
        recipe.meta["updated"] = time.time()
        recipe.save()

        recipe.load()

    def _delete(self, template: str) -> None:
        """Delete a project template from this repository."""
        tpl_dir = self._root / template
        if tpl_dir.is_dir():
            if tpl_dir.is_symlink():
                tpl_dir.unlink()
            else:
                shutil.rmtree(tpl_dir)

    def _reload(self) -> t.List[str]:
        diff = list()
        for child in self._root.iterdir():
            if child.is_dir():
                project_file = child / PRJ_FILE
                if project_file.is_file():
                    if child.name not in self._recipes:
                        diff.append(child.name)
                        self._recipes.append(child.name)
        self._recipes.sort()
        return diff


@dataclass
class Boiler:
    """A `Boiler` renders a [Recipe][parboil.recipes.Recipe] into a `target_dir`.

    In this process the user may be is prompted for answers to the fields
    configured in the recipes config file.

    Raises:
        TaskFailedError: If a task exited with a returncode other than zero.
        TaskExecutionError: If a task fails execution.

    Attributes:
        recipe:
            The recipe this `Boiler` will render.
        target_dir:
            Directory the recipe is renderered into.
        prefilled:
            A dicttionary with already defined variable values.
        context:
            A dictionary with context variables to use for template rendering.
    """

    recipe: Recipe
    target_dir: Path

    prefilled: t.Dict[str, t.Any]
    context: t.ChainMap[str, t.Any] = field(default_factory=ChainMap)

    def fill(self) -> None:
        """
        Get field values either from the prefilled values or read user input.
        """
        for _field in self.recipe.ingredients:
            self.renderer.render_obj(_field, INGREDIENT=_field)

            if not eval_bool(_field.condition or True):
                console.info(
                    f'Skipped field "[ingredient]{_field.name}[/]" due to failed condition'
                )
                continue
            elif _field.name in self.prefilled:
                self.context[_field.name] = _field.value = self.renderer.render_string(
                    self.prefilled[_field.name], INGREDIENT=_field
                )
                console.info(f'Used prefilled value for "[ingredient]{_field.name}[/]"')
            else:
                self.context[_field.name] = _field.prompt(self)

        for key, descr in self.recipe.context.items():
            self.context[key] = self.renderer.render_string(descr)

    def compile(self) -> t.Generator[t.Tuple[bool, Path, t.Optional[Path]], None, None]:
        """Compile the recipe into the target directory.

        Attempts to compile every file in `self.templates` with [jinja2](#) and to save it to its final location in the target directory.

        Yields a tuple with three values for each template file:



        Yields:
            (bool, str, str): A tuple holding

                1. `True`, if an output file was generated, `False` otherwise,
                2. the original file.
                3. The output file after compilation or `None`, if no file was rendered.
        """
        ## Create target directory
        self.target_dir.mkdir(parents=True, exist_ok=True)

        ## Execute pre-run tasks
        self.execute_tasks("pre-run")

        # TODO Error handling
        for _file in self.recipe.templates:
            if isinstance(_file, Recipe):
                # TODO refactor subproject inclusion (field and compilation to tighly coupled)
                subproject = Boiler(_file, self.target_dir, self.prefilled)
                yield from subproject.compile()
            else:
                _file = Path(_file)
                file_in = Path(str(_file).removeprefix("includes:"))
                file_out = str(file_in)
                file_cfg: t.Dict[str, t.Any] = self.recipe.files.get(
                    str(file_in), dict()
                )
                file_out = file_cfg.get("filename", file_out)

                rel_path = file_in.parent
                abs_path = self.target_dir / rel_path

                # Set some dynamic values
                boil_vars = dict(
                    RELDIR="" if rel_path.name == "" else str(rel_path),
                    ABSDIR=str(abs_path),
                    OUTDIR=str(self.target_dir),
                    OUTNAME=str(self.target_dir.name),
                )
                path_render = self.renderer.render_string(file_out, BOIL=boil_vars)

                if Path(path_render).exists() and not file_cfg.get("overwrite", True):
                    yield (False, file_in, None)
                    continue

                boil_vars["FILENAME"] = Path(path_render).name
                boil_vars["FILEPATH"] = path_render

                if file_cfg.get("render", True):
                    # Render template
                    tpl_render = self.renderer.render_file(_file, BOIL=boil_vars)
                else:
                    tpl_render = self.recipe.templates_dir.joinpath(_file).read_text()

                generate_file = bool(tpl_render.strip())  # empty?
                generate_file = file_cfg.get("keep", generate_file)

                if generate_file:
                    path_render_abs = self.target_dir / path_render
                    path_render_abs.parent.mkdir(parents=True, exist_ok=True)
                    path_render_abs.write_text(tpl_render)

                    yield (True, _file, Path(path_render))
                else:
                    yield (False, _file, Path(path_render))

        # Execute post-run tasks
        self.execute_tasks("post-run")

    def execute_tasks(self, hook: str) -> None:
        if hook not in self.recipe.tasks:
            return

        total_tasks = len(self.recipe.tasks[hook])
        with self.cwd():
            for i, task in enumerate(self.recipe.tasks[hook]):
                self.renderer.render_obj(task, TASK=task)
                console.info(
                    f"Running [keyword]{hook}[/] task {i+1} of {total_tasks}: [cmd]{task}[/]"
                )
                try:
                    if not task.execute():
                        raise TaskFailedError(task)
                except Exception as e:
                    raise TaskExecutionError(task) from e

    @cached_property
    def renderer(self) -> ParboilRenderer:
        return ParboilRenderer(self)

    @contextmanager
    def cwd(self) -> t.Iterator[Path]:
        """Change working dir to `self.target_dir` for task execution."""
        _current = Path.cwd()
        os.chdir(self.target_dir)
        yield self.target_dir
        os.chdir(_current)



# -*- coding: utf-8 -*-
"""
Initialisation of Jinja2 environment and rendering of templates
from files or strings.
"""


import os
import sys
from collections.abc import MutableSequence
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Generator, Protocol, Union

import jinja2_ansible_filters
from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PrefixLoader
from jinja2 import Template as JinjaTemplate
from jinja2.sandbox import SandboxedEnvironment
from rich import inspect

from .ext import jinja_filter_fileify, jinja_filter_roman, jinja_filter_slugify

if TYPE_CHECKING:
    from parboil.recipe import Boiler


class ParboilRenderable(Protocol):
    """Protocol for classes that can be rendered by [parboil.renderer.ParboilRenderer.render_obj()][]."""

    def __templates__(self) -> Generator[str, str, None]:
        ...


def renderable(cls=None, *attrs, strict: bool = True, render_empty: bool = False):
    """Decorator to make a class a ParboilRenderable."""

    def wrapper(cls):
        if not hasattr(cls, "__templates__"):

            def _render(self) -> Generator[str, str, None]:
                for key in attrs:
                    if hasattr(self, key):
                        val = getattr(self, key, None)
                        if isinstance(val, str):
                            setattr(self, key, (yield val))
                        elif not strict:
                            setattr(self, key, (yield str(val)))
                    elif render_empty:
                        # set attr directly to empty string?
                        setattr(self, key, (yield ""))

            setattr(cls, "__templates__", _render)
        return cls

    if cls is None:
        return wrapper
    return wrapper(cls)


# TODO Exception handling
class ParboilRenderer:
    def __init__(self, boiler: "Boiler"):
        self._boiler = boiler
        self._environ = os.environ.copy()

    @cached_property
    def env(self) -> Environment:
        """Creates a jinja Environment for this project and caches it"""
        env = SandboxedEnvironment(
            loader=ChoiceLoader(
                [
                    FileSystemLoader(self._boiler.recipe.templates_dir),
                    PrefixLoader(
                        {
                            "includes": FileSystemLoader(
                                self._boiler.recipe.includes_dir
                            )
                        },
                        delimiter=":",
                    ),
                ]
            ),
            extensions=[jinja2_ansible_filters.AnsibleCoreFiltersExtension],
        )
        env.filters["fileify"] = jinja_filter_fileify
        env.filters["slugify"] = jinja_filter_slugify
        env.filters["roman"] = jinja_filter_roman

        return env

    def _render_template(self, template: JinjaTemplate, **kwargs) -> str:
        if "BOIL" not in kwargs:
            kwargs["BOIL"] = dict()
        kwargs["BOIL"]["TPLNAME"] = self._boiler.recipe.name
        kwargs["BOIL"]["RUNTIME"] = sys.executable

        return template.render(
            **self._boiler.context,
            **kwargs,
            ENV=self._environ,
            BOILER=self._boiler,
            RECIPE=self._boiler.recipe
        )

    def render_string(self, template: str, **kwargs) -> str:
        return self._render_template(self.env.from_string(str(template)), **kwargs)

    def render_strings(
        self, templates: MutableSequence[str], **kwargs
    ) -> MutableSequence[str]:
        """Render a sequence of string templates in place."""
        for i, tpl in enumerate(templates):
            templates[i] = self.render_string(tpl, **kwargs)
        return templates

    def render_obj(self, renderable: ParboilRenderable, **kwargs):
        templates = renderable.__templates__()
        try:
            # start generator function
            template = next(templates)
            while True:
                rendered = self.render_string(template, **kwargs)
                template = templates.send(rendered)
        except StopIteration:
            templates.close()

    def render_file(
        self, filename: Union[str, Path], render_filename: bool = False, **kwargs
    ) -> str:
        """Renders a file as jinja2 template.

        If `render_filename`is `True`, the `filename` will be rendered with `render_string` first, before attempting to load the template file."""
        if render_filename:
            filename = self.render_string(str(filename), **kwargs)
        else:
            filename = str(filename)

        return self._render_template(self.env.get_template(filename))



"""Collection of common errors."""


class ParboilError(Exception):
    pass


class RecipeError(ParboilError):
    pass


class ProjectError(ParboilError):
    pass


class ProjectFileNotFoundError(ProjectError):
    def __init__(self, msg="Project file not found."):
        super().__init__(msg)


class ProjectExistsError(ProjectError):
    def __init__(self, msg="Project file already exists."):
        super().__init__(msg)


class ProjectConfigError(ProjectError):
    pass


class RepositoryError(ParboilError):
    def __init__(self, msg, repository):
        self.repository = repository


class RecipeNotInstalledError(RepositoryError):
    def __init__(self, template, repository):
        self.repository = repository
        super().__init__(f"Project {template} not installed.")


class BoilerError(ParboilError):
    pass


class TaskExecutionError(ParboilError):
    def __init__(self, task, msg: str = None):
        self.task = task

        if not msg:
            msg = f"Error executing task: <{task.quoted()}>"
        super().__init__(msg)


class TaskFailedError(ParboilError):
    def __init__(self, task, msg: str = None):
        self.task = task

        if not msg:
            msg = f"Task exited with error code {task.returncode}: <{task.quoted()}>"
        super().__init__(msg)
